package com.ZoS.RoyalFlush;

import java.awt.Font;

import org.lwjgl.Sys;
import org.lwjgl.opengl.GL11;
import org.newdawn.slick.Color;
import org.newdawn.slick.TrueTypeFont;

import com.ZoS.RoyalFlush.GUI.OGLDisplay;
import com.ZoS.RoyalFlush.GUI.Window;
import com.ZoS.RoyalFlush.Game.IGame;
import com.ZoS.RoyalFlush.Input.Keyboard;
import com.ZoS.RoyalFlush.Input.MouseCursor;
import com.ZoS.RoyalFlush.Models.AxisGuide;
import com.ZoS.RoyalFlush.Renderer.OGLRenderer;

//TODO: Remove getWindow
//TODO: Add hooks for user to get information
// - IsFocus, IsVisible, Resize, etc
public final class RoyalFlushApplication {
	//The main game display
	private OGLDisplay display		= null;
	private OGLRenderer renderer	= null;
	
	private MouseCursor mouse		= null;
	private Keyboard keyboard		= null;
	
	//Frames 
	private int FPS = 0;
	private int frameCount = 0;
	private long lastFPS = 0;
	private long timeLastFrame = 0;
	
	//Debug variables
	private boolean debugMenu = true;
	private TrueTypeFont debugFont;
	
	//Game Variables
	private IGame game = null;
	private boolean running = false;
	private String[] commandLineArgs = null;
	
	/**
	 * Constructor: Creates the main game engine.
	 * 
	 * @param _game - The game to run that extends IGame.
	 * @param args - Arguments passed in from main().
	 */
	public RoyalFlushApplication(IGame _game, String[] args) {
		//TODO: Check to see if _game is NULL
		game = _game;
		commandLineArgs = args;
	}
	
	/**
	 * Functions for the user
	 */
	public void startApplication(int _width, int _height, boolean _fullscreen, boolean _vsync, String _title) {
		
		//First make sure we have a game to work with
		if(this.game == null) {
			this.stopApplication("No game has been specified to run!");
		}
		
		display = new OGLDisplay(_width, _height, _fullscreen, _vsync, _title);
		parseCommandLineArgs(commandLineArgs);
		
		renderer = new OGLRenderer(this.display.getWindow());
		mouse = new MouseCursor(this.display.getWindow());
		keyboard = new Keyboard();
		
		//Initialize the game engine.
		this.Initialize();
		
		//Initialize the game
		if(!this.game.Initialize()) {
			this.stopApplication(this.game.getClass().getSimpleName() + " failed to initalize!");
		}
		
		//Now that everything is ready to go begin running the application.
		running = true;
		
		//TODO Make this its own thread?
		while (!this.display.isCloseRequested() && this.running) {
			if(!gameLoop()) {
				running = false;
			}
		}
		
		//If we make it here the application is done running, time to shutdown.
		this.stopGame();
		this.stopApplication();
	}
	
	/**
	 * Initialize everything the engine needs. 
	 * @return
	 */
	private void Initialize() {
		this.display.Create();
		this.renderer.initGL();
		
		getDelta();
		this.lastFPS = getTime();
		
		Font awtFont = new Font("Lucida Console", Font.PLAIN, 16);
		this.debugFont = new TrueTypeFont(awtFont, true);
	}
	
	/**
	 * Private Functions
	 */
	private void stopGame() {
		this.game.Shutdown();
		this.display.Shutdown();
		running = false;
	}
	
	/**
	 * Get the time in milliseconds
	 * 
	 * @return The system time in milliseconds
	 */
	private long getTime() {
	    return (Sys.getTime() * 1000) / Sys.getTimerResolution();
	}
	
	private double getDelta() {
	    long time = getTime();
	    double delta = (double) (time - timeLastFrame);
	    timeLastFrame = time;
	    	
	    return delta;
	}
	
	private void updateFPS() {
	    if (getTime() - lastFPS > 1000) {
	    	FPS = frameCount;
	        frameCount = 0; //reset the FPS counter
	        lastFPS += 1000; //add one second
	    }
	    frameCount++;
	}
	
	private void stopApplication() {
		running = false;
		System.out.print("RoyalFlush: Game shutdown, thanks for playing!\n");
		System.exit(0);
	}
	
	private void stopApplication(String _errorMessage) {
		running = false;
		System.err.print("RoyalFlush: " + _errorMessage + "\n");
		System.exit(0);
	}
	
	private boolean gameLoop() {
		
		updateFPS();
		double deltaTime = this.getDelta();
		
		if(!game.Input(deltaTime, mouse, keyboard)) {	//If the Input fails, return false
			return false;
		}
		
		if(!game.Update(deltaTime)) {	//If the Update fails, return false
			return false;
		}
		
		renderer.glSetup3D();
		GL11.glPushMatrix();
		{
			//Temp translate to see guide
			GL11.glTranslatef(0.0f, -3.0f, -10.0f);
			AxisGuide.draw3DAxisGuide();
		
			if(!game.Render3D(deltaTime)) {	//If the Render fails, return false
				return false;
			}
		}
		GL11.glPopMatrix();
		
		renderer.glSetup2D();
		GL11.glPushMatrix();
		{
			if(!game.Render2D(deltaTime)) {	//If the Render fails, return false
				return false;
			}
		}
		GL11.glPopMatrix();
		
		if(debugMenu) {
			GL11.glPushMatrix();
			
			GL11.glEnable(GL11.GL_BLEND);
			GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
			
			GL11.glTranslatef(0.0f, 0.0f, -1.0f);
			
			int YtoDraw = 0;
			
			//Render debug information
			debugFont.drawString(0, YtoDraw, "Debug Menu", Color.white);
			YtoDraw += debugFont.getLineHeight();
			debugFont.drawString(0, YtoDraw, " - FPS: " + FPS + " (" + frameCount + ")", Color.white);
			YtoDraw += debugFont.getLineHeight();
			debugFont.drawString(0, YtoDraw, " - Mouse X: " + this.mouse.getX(), Color.white);
			YtoDraw += debugFont.getLineHeight();
			debugFont.drawString(0, YtoDraw, " - Mouse Y: " + this.mouse.getY(), Color.white);
			
			GL11.glDisable(GL11.GL_BLEND);
			GL11.glPopMatrix();
		}
		
		this.display.Update();
		return true;
	}
	
	public void toggleDebugText() {
		debugMenu = !debugMenu;
	}
	
	private enum CommandLineArgs {
		Width, Height, Fullscreen, VSync,
		NOVALUE;
		
		public static CommandLineArgs toCommand(String _string)
	    {
	        try {
	            return valueOf(_string);
	        } 
	        catch (Exception ex) {
	            return NOVALUE;
	        }
	    } 
	}
	
	private void parseCommandLineArgs(String[] _args) {
		//Process String[] args for passed in parameters
		// --rfWidth, --rfHeight, --rfFullscreen, --rfVSync
		for(String s: _args) {
			
			//Split the command line arguments into a key and value
			String[] command = s.split("=");
			
			//There should be a key and a value. If there is more then that
			// (more then 1 '=') then the command is invalid.
			if(command.length == 2) {
				
				//Split the key to see if its an engine command.
				String[] key = command[0].split("--rf");
				
				//Make sure the key split into 2 halves. If it did then it is a
				// valid engine command.
				if(key.length == 2) {

					switch(CommandLineArgs.toCommand(key[1])) {
						case Width: {
							//TODO: Too much power? Error check?
							try {
								this.display.setWidth(Integer.parseInt(command[1]));
								System.out.print("RF Command: " + key[1] + " set to value: " + command[1] + "\n");
							}
							catch(NumberFormatException e) {
								System.err.print("RF Command: " + key[1] + " cannot set to value: " + command[1] + "\n");
							}
							break;
						}
						
						case Height: {
							try {
								this.display.setHeight(Integer.parseInt(command[1]));
								System.out.print("RF Command: " + key[1] + " set to value: " + command[1] + "\n");
							}
							catch(NumberFormatException e) {
								System.err.print("RF Command: " + key[1] + " cannot set to value: " + command[1] + "\n");
							}
							break;
						}
						
						case Fullscreen: {
							this.display.setFullScreen(Boolean.parseBoolean(command[1]));
							System.out.print("RF Command: " + key[1] + " set to value: " + command[1] + "\n");
							break;
						}
						
						case VSync: {
							this.display.setFullScreen(Boolean.parseBoolean(command[1]));
							System.out.print("RF Command: " + key[1] + " set to value: " + command[1] + "\n");
							break;
						}
					
						default: {
							System.err.print("RF Command: " + s + " is not supported.\n");
							break;
						}
					}
				}
				else {
					System.err.print("RF Command: " + s + " skipped. Not an engine command.\n");
				}
			}
			else {
				//TODO: Log this.
				System.err.print("RF Command: " + s + " failed to parse.\n");
			}
		}
	}

	/**
	 * Display Wrapper Functions:
	 *  - Allow the user to access the display and change all the settings
	 *    without giving them full access to everything and the application breaking.
	 */
	
	public Window getWindow() {
		return this.display.getWindow();
	}
	
	public boolean isFocused() {
		return this.display.isFocused();
	}
	
	public boolean isVisible() {
		return this.display.isVisible();
	}
	
	public void centerApplication() {
		this.display.centerApplication();
	}
	
	public void setWidth(int _width) {
		this.display.setWidth(_width);
	}
	
	public void setHeight(int _height) {
		this.display.setHeight(_height);
	}
	
	public void setWindowSize(int _width, int _height) {
		this.display.setWindowSize(_width, _height);
	}
	
	public void setX(int _x) {
		this.display.setX(_x);
	}
	
	public void setY(int _y) {
		this.display.setY(_y);
	}
	
	public void setLocation(int _x, int _y) {
		this.display.setLocation(_x, _y);
	}
	
	public void setFullscreen(boolean _fullscreen) {
		this.display.setFullScreen(_fullscreen);
	}
	
	public void toggleFullscreen() {
		this.display.toggleFullScreen();
	}
	
	public void setVSync(boolean _vsync) {
		this.display.setVSync(_vsync);
	}
	
	public void toggleVSync() {
		this.display.toggleVSync();
	}
	
	public void setResizable(boolean _resizable) {
		this.display.setResizable(_resizable);
	}
	
	public void toggleResizable() {
		this.display.toggleResizable();
	}
	
	public void setTitle(String _title) {
		this.display.setTitle(_title);
	}
	
	public void setMaxFPS(int _maxFPS) {
		this.display.setMaxFPS(_maxFPS);
	}
}