package com.ZoS.RoyalFlush.Texture;

import java.io.IOException;

import org.lwjgl.opengl.GL11;
import org.lwjgl.util.vector.Vector2f;
import org.newdawn.slick.opengl.Texture;
import org.newdawn.slick.opengl.TextureLoader;
import org.newdawn.slick.util.ResourceLoader;

/**
 * A wrapper for the slick texture to streamline the loading
 *  of textures and aid in correcting UV's on non-power-of-two images.
 *  
 * @author JpDeathBlade
 */
public class Texture2D {
	
	private Texture	texture;		// The actual texture we loaded.
	private String	fileName;		// The name of the texture we loaded.
	private boolean	isBound;		// True if the texture is bound.
	private int		prevTexture;	// If bound, holds the value of the previous texture, otherwise -1.
	
	private float widthRatio;
	private float heightRatio;
	
	/**
	 * Constructor: Loads in the texture file passed in. If the image is not
	 *  a power of two it will fill in the blank areas with black. It is suggested
	 *  you make your image a power of two always.
	 *  
	 * @param _format - The format of the image.
	 * @param _file - The path to the file to be loaded.
	 */
	//TODO: Create init function, nothing that can fail should be called in constructor
	public Texture2D(TextureFormat _format, String _file) {
		super();
		
		isBound = false;
		prevTexture = -1;
		widthRatio = 0.0f;
		heightRatio = 0.0f;
		
		if(!LoadImage(_format, _file)) {
			//TODO: Default to error texture
		}
	}
	
	public int getTextureID(){
		if(texture != null) {
			return texture.getTextureID();
		}
		else {
			System.err.print("RoyalFlush: Texture2D.getTextureID - No texture has been loaded.\n");
			return -1;
		}
	}
	
	public void bind() {
		if(texture != null && !isBound) {
			//Save the previous texture to set back, otherwise there's issues with fonts
			prevTexture = GL11.glGetInteger(GL11.GL_TEXTURE_BINDING_2D);
			GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture.getTextureID());
			isBound = true;
		}
		else if(texture != null && isBound) {
			System.err.print("RoyalFlush: Texture2D.bind - Texture " + fileName + " already bound.\n");
		}
		else {
			System.err.print("RoyalFlush: Texture2D.bind - No texture has been loaded.\n");
		}
	}
	
	public void unbind() {
		//Make sure there was a previous texture before we try to unbind.
		if(prevTexture != -1 && isBound) {
			GL11.glBindTexture(GL11.GL_TEXTURE_2D, prevTexture);
			prevTexture = -1;
			isBound = false;
		}
		else if(prevTexture != -1 && !isBound) {
			System.err.print("RoyalFlush: Texture2D.unbind - Texture " + fileName + " is NOT bound.\n");
		}
		else {
			System.err.print("RoyalFlush: Texture2D.unbind - No texture to unbind.\n");
		}
	}
	
	public float getImageHeight() {
		if(texture != null) {
			return (float)texture.getImageHeight();
		}
		else {
			System.err.print("RoyalFlush: Texture2D.getImageHeight - No texture has been loaded.\n");
			return -1.0f;
		}
	}
	
	public float getImageWidth() {
		if(texture != null) {
			return (float)texture.getImageWidth();
		}
		else {
			System.err.print("RoyalFlush: Texture2D.getImageWidth - No texture has been loaded.\n");
			return -1.0f;
		}
	}
	
	public float getTextureHeight() {
		if(texture != null) {
			return (float)texture.getTextureHeight();
		}
		else {
			System.err.print("RoyalFlush: Texture2D.getTextureHeight - No texture has been loaded.\n");
			return -1.0f;
		}
	}
	
	public float getTextureWidth() {
		if(texture != null) {
			return (float)texture.getTextureWidth();
		}
		else {
			System.err.print("RoyalFlush: Texture2D.getTextureWidth - No texture has been loaded.\n");
			return -1.0f;
		}
	}
	
	/**
	 * If the image isn't a power of two the texture library will make it one so
	 *  your UV's will be off. This function will correct the U value.
	 *  
	 * @param _u
	 * @return
	 */
	public float UtoTextureSpace(float _u) {
		if(texture != null) {
			return (widthRatio * _u);
		}
		else {
			System.err.print("RoyalFlush: Texture2D.UtoTextureSpace - No texture has been loaded.\n");
			return _u;
		}
	}
	
	/**
	 * If the image isn't a power of two the texture library will make it one so
	 *  your UV's will be off. This function will correct the V value.
	 *  
	 * @param _v
	 * @return
	 */
	public float VtoTextureSpace(float _v) {
		if(texture != null) {
			return (heightRatio * _v);
		}
		else {
			System.err.print("RoyalFlush: Texture2D.VtoTextureSpace - No texture has been loaded.\n");
			return _v;
		}
	}
	
	/**
	 * If the image isn't a power of two the texture library will make it one so
	 *  your UV's will be off. This function will correct the UV values.
	 * 
	 * @param _u
	 * @param _v
	 * @return
	 */
	public Vector2f UVtoTextureSpace(float _u, float _v) {
		float u = this.UtoTextureSpace(_u);
		float v = this.VtoTextureSpace(_v);
		return new Vector2f(u, v);
	}
	
	/**
	 * If the image isn't a power of two the texture library will make it one so
	 *  your UV's will be off. This function will correct the UV values.
	 * 
	 * @param _uv
	 * @return
	 */
	public Vector2f UVtoTextureSpace(Vector2f _uv) {
		return UVtoTextureSpace(_uv.x, _uv.y);
	}
	
	/**
	 * 
	 * @param _format
	 * @param _file
	 * @return
	 */
	private boolean LoadImage(TextureFormat _format, String _file) {
		fileName = _file;
		String _texFormat = getFormatString(_format);

		try {
			texture = TextureLoader.getTexture(_texFormat, ResourceLoader.getResourceAsStream(_file));
		} catch (IOException e) {
			System.err.print("RoyalFlush: Texture2D.LoadImage - Failed to load texture " + fileName + "\n");
			return false;
		}
		
		widthRatio = this.getImageWidth() / this.getTextureWidth();
		heightRatio = this.getImageHeight() / this.getTextureHeight();
		
		return true;
	}
	
	/**
	 * Returns the proper string based on the image format.
	 * TODO: Rewrite to scan file extension.
	 *       Support more variations (Like JPEG).
	 * 
	 * @param _format - Format the image is in.
	 * @return String - value for the given image format.
	 */
	private String getFormatString(TextureFormat _format) {
		switch(_format){
			case PNG: {
				return "PNG";
			}
			case JPG: {
				return "JPG";
			}
			case GIF: {
				return "GIF";
			}
			case TGA: {
				return "TGA";
			}
			default: {
				//TODO: Error, texture type not supported. Return error texture.
				return"PNG";
			}
		}
	}
}